"""Monkey-patch FileIOCalculator to enable async calculation.
Some other opionionate tweaks are included.
"""
from ase.calculators.calculator import *
import trio


# Function call cascade (take Gaussian as an example):
# Gaussian/Calc.get_property ->
# Gaussian.calculate(atoms, [name], system_changes) ===
# Gaussian.calculate(Gaussian.atoms, [name], []) "given no system_changes"

# The original method was to ensure `self.directory` exists on fs, which we don't care.
def FileIOCalculator__write_input(self, atoms, properties=None, system_changes=None):
    pass

def FileIOCalculator__calculate(self, *args, **kwargs):
    raise TypeError(
         "FileIOCalculator's sync calculate function should never be triggered. "
        f"Call `await self.acalculate({', '.join(map(str,args))}, {', '.join(str(k)+'='+str(v) for k,v in kwargs)})` "
         "before accessing any calculation result/property."
    )

# adapted from https://gitlab.com/ase/ase/-/blob/02c57173e032d31dd88aea005a3787475d017752/ase/calculators/calculator.py
async def FileIOCalculator__acalculate(self, atoms=None, properties=['energy'],
                                      system_changes=all_changes):
    Calculator.calculate(self, atoms, properties, system_changes)
    self.write_input(self.atoms, properties, system_changes)
    if self.command is None:
        raise CalculatorSetupError(
            'Please set ${} environment variable '
            .format('ASE_' + self.name.upper() + '_COMMAND') +
            'or supply the command keyword')
    command = self.command
    if 'PREFIX' in command:
        command = command.replace('PREFIX', self.prefix)

    try:
        #proc = subprocess.Popen(command, shell=True, cwd=self.directory)
        cproc = await trio.run_process(command, check=False, shell=True, cwd=self.directory)
    except OSError as err:
        # Actually this may never happen with shell=True, since
        # probably the shell launches successfully.  But we soon want
        # to allow calling the subprocess directly, and then this
        # distinction (failed to launch vs failed to run) is useful.
        msg = 'Failed to execute "{}"'.format(command)
        raise EnvironmentError(msg) from err

    #errorcode = proc.wait()
    errorcode = cproc.returncode

    if errorcode:
        # NOTE: we don't care about directory
        #path = os.path.abspath(self.directory)
        #msg = ('Calculator "{}" failed with command "{}" failed in '
        #       '{} with error code {}'.format(self.name, command,
        #                                      path, errorcode))
        msg = (f'Calculator "{self.name}" failed with command "{command}" '
               f'failed with error code {errorcode}')
        raise CalculationFailed(msg)

    self.read_results()

# get rid of the annoying .ase parameter file
def Parameters__write(self, filename):
    pass

# Make set_psi4 executed only once
def Psi4__init(self, *args, **kwargs):
    Calculator.__init__(self, *args, **kwargs)
    import psi4
    self.psi4 = psi4
    # perform initial setup of psi4 python API
    #self.set_psi4(atoms=atoms)


import ase
ase.calculators.calculator.FileIOCalculator.write_input = FileIOCalculator__write_input
ase.calculators.calculator.FileIOCalculator.calculate = FileIOCalculator__calculate
ase.calculators.calculator.FileIOCalculator.acalculate = FileIOCalculator__acalculate
ase.calculators.calculator.Parameters.write = Parameters__write
import ase.calculators.psi4
ase.calculators.psi4.Psi4.__init__ = Psi4__init



from ase.data import atomic_masses

from contextlib import ContextDecorator
from timeit import default_timer as timer

def timeblock(text="this block", enter=False):
    class TimeblockCtx(ContextDecorator):
        def __enter__(self):
            self.st = timer()
            if enter:
                print(f"[{self.st:.3f}]\tEnter {text}")
            return self
        def __exit__(self, *exc):
            et = timer()
            print(f"[{et:.3f}]\tTime elapsed for {text}{' (incomplete)' if any(exc) else ''}: {et-self.st}")
            return False
    return TimeblockCtx()

def water_charge(ian):
    """Calculate charge for a water cluster's fragment.
    Assign +1 for H and -2 for O.

    Args:
        ian (ndarray[(N,), int]): Atomic numbers

    Returns:
        charge in int
    """
    return sum({ 1:+1, 8:-2 }[i] for i in ian)

def center_of_mass(acoords, ian):
    """Calculate the center of mass for a group of atoms.

    Args:
        acoords (ndarray[(N,3), float]): Atoms' coordinates
        ian (ndarray[(N,), int]): Atomic numbers

    Returns:
        coordinate for the center of mass
        (ndarray[(3,), float])
    """
    m = atomic_masses[ian]
    return m @ acoords / sum(m)



from grain.subproc import subprocify

async def ase_fio_task(cid, calc, atcor, ian, cell=None, pbc=None):
    """Base task for calculators implementing the
    `FileIOCalculator`_ interface (e.g. Gaussian,
    Orca, QE).

    .. _FileIOCalculator: https://gitlab.com/ase/ase/-/blob/master/ase/calculators/calculator.py

    Class ``FileIOCalculator`` is patched to enable
    async calculation through async subprocess. See
    ``monkey_patch.py`` for details.

    Args:
      cid (str): Calculation ID, for time record only
      calc: The calculator instance
      atcor (ndarray[(N,3), float]): Atom coordinate
      ian (ndarray[(N,), int]): Atomic numbers 
      pbc (bool or (bool, bool, bool)): Periodic
        boundary conditions along each axis. Arg for
        ``ase.atoms.Atoms``
      cell (ndarray[(3,3), float] or see ASE doc):
        Unit cell size. Arg for ``ase.atoms.Atoms``

    Returns:
      Energy (in hartree) and forces (in hartree/angstrom)

    .. note:: Any task calling this should make
       sure that the calculator itself's ``calculate``
       method get handled, as this only takes care
       of ``FileIOCalculator.calculate``.
    """
    from ase.atoms import Atoms
    from ase.units import Hartree as hart2ev, Bohr as b2a
    from .util import timeblock
    mol = Atoms(numbers=ian, positions=atcor, cell=cell, pbc=pbc)
    mol.calc = calc
    with timeblock(cid):
        await mol._calc.acalculate(mol, ['energy', 'forces'], [])
    return (
        mol.get_potential_energy() / hart2ev,
        mol.get_forces() * b2a / hart2ev,
    )

async def dumb_fio_task(cid, calc, atcor, ian, cell=None, pbc=None):
    import numpy as np
    from ase.atoms import Atoms
    mol = Atoms(numbers=ian, positions=atcor, cell=cell, pbc=pbc)
    mol.calc = calc
    mol._calc.write_input(mol, ['energy', 'forces'], [])
    return (
        0.0,
        np.zeros_like(atcor),
    )

@subprocify
def ase_task(cid, calc, atcor, ian):
    """Base task for inproc calculators (e.g. Psi4).

    For Parameters and Returns see above :func:`ase_fio_task`.

    :func:`ase_task` uses ``grain.subproc.subprocify``
    to "asyncify" the calculator, as we assume that an
    inproc calculator's ``calculate`` method is CPU
    intensive. ``subprocify`` requires an async
    context to maintain a subprocess pool. This could
    be simply achieved by::

      from grain.subproc import subprocess_pool_scope
      async with subprocess_pool_scope():
          # Safely call subprocified functions within

    .. note:: Task calling this experiences an overhead
       for the first few runs due to ``subprocify``'s
       subprocess startup cost, which will be amortized
       later as the subprocesses are reused.
    """
    from ase.atoms import Atoms
    from ase.units import Hartree as hart2ev, Bohr as b2a
    from .util import timeblock
    mol = Atoms(numbers=ian, positions=atcor)
    mol.calc = calc
    with timeblock(cid):
        mol._calc.calculate(mol, ['energy', 'forces'], [])
    return (
        mol.get_potential_energy() / hart2ev,
        mol.get_forces() * b2a / hart2ev,
    )



from . import monkey_patch
from .base import ase_fio_task, ase_task, dumb_fio_task

from ase.calculators.gaussian import Gaussian
from ase.calculators.psi4 import Psi4
from ase.calculators.espresso import Espresso
from ase.calculators.orca import ORCA

from grain import GVAR

from os import environ

async def gautask(cid, mb, chg_mlt, atcor, ian, cell=None, pbc=None, gau="g16", scrmem=0, **kwargs):
    """
    Args:
      cell (ndarray[(3,3), float] or see ASE doc):
        Unit cell size. Arg for ``ase.atoms.Atoms``
      pbc (bool or (bool, bool, bool)): Periodic
        boundary conditions along each axis. Arg for
        ``ase.atoms.Atoms``
      gau (str): the Gaussian executable
      scrmem (int): memory (in GB), out of the total
        memory allocated, for scratch dir on a memory
        fs.

    Any other keyword args are treated as Link0 or
    route keywords. See ASE's `Gaussian`_ calculator
    for reference.

    .. _Gaussian: https://wiki.fysik.dtu.dk/ase/ase/calculators/gaussian.html#parameters
    """
    mb = mb.split('/')
    ioplist = None
    if len(ian) > 50:
        ioplist = ('2/9=2000',) # Gaussian calculator relies on 'input orientation' block. Turn it on.
    calc = Gaussian(
        **__dl(cid), method=mb[0], basis=mb[1] if len(mb)>1 else None,
        charge=chg_mlt[0], mult=chg_mlt[1],
        cpu=__cs(GVAR.res.c),
        mem=f"{GVAR.res.m-scrmem}GB",
        ioplist=ioplist,
        **kwargs,
    )
    # This is done in Gaussian.calculate, but we skip it due to the patched async calculation
    calc.command = calc.command.replace('GAUSSIAN', gau)
    return await ase_fio_task(cid, calc, atcor, ian, cell, pbc)

async def psi4task(cid, mb, chg_mlt, atcor, ian):
    """:func:`psi4task` is implemented using
    ``grain.subproc.subprocify`` For notes and caveats
    refer to the base task :func:`ase_grain.base.ase_task`
    or ``subprocify``'s documentation.
    """
    m, b = mb.split('/')
    calc = Psi4(
        **__dl(cid), method=m, basis=b,
        charge=chg_mlt[0], multiplicity=chg_mlt[1],
        num_threads=GVAR.res.N, # use N threads for N cores, and subprocify handles which cores to bind
        memory=f"{getattr(GVAR.res,'m',0.5)}GB",
    )
    return await ase_task(cid, calc, atcor, ian)

async def qetask(cid, atcor, ian, cell,
                 pseudopotentials, kspacing=None, kpts=None, koffset=(0,0,0),
                 input_data=None, mpi="Open MPI", **kwargs):
    """
    QE is very different from other backends. See ASE's
    `Espresso`_ calculator for reference.

    .. _Espresso: https://wiki.fysik.dtu.dk/ase/ase/calculators/espresso.html#ase.calculators.espresso.Espresso

    Args:
        cell (ndarray[(3,3), float]): Unit cell vectors.
            This is passed to ``ase.atoms.Atoms``
        mpi (str): Whether to use Open MPI or Intel MPI

    .. note:: Currently there is no way to impose a
       memory constraint on QE. Please measure ahead
       the memory consumption for your calculation, and
       use that value to inform the scheduler. There is
       no guarentee for QE to go over that value, so you
       might want to set a higher one.
    """
    input_data = input_data or {}
    input_data['tprnfor'] = True # print forces
    calc = Espresso(
        **__dl(cid), input_data=input_data,
        pseudopotentials=pseudopotentials,
        kspacing=kspacing, kpts=kpts, koffset=koffset,
        **kwargs,
    )
    cpu = __cs(GVAR.res.c)
    if mpi == "Open MPI":
        para_prefix = f'mpirun -np {GVAR.res.N} --bind-to cpulist:ordered --cpu-set {cpu} '
    elif mpi == "Intel MPI":
        para_prefix = f'mpirun -np {GVAR.res.N} -genv I_MPI_PIN_PROCESSOR_LIST={cpu} '
    else:
        raise TypeError(f"Unsupported MPI type: {mpi}")
    calc.command = para_prefix + 'pw.x -in PREFIX.pwi > PREFIX.pwo'
    return await ase_fio_task(cid, calc, atcor, ian, cell)

async def orcatask(cid, mb, chg_mlt, atcor, ian, orca="", simple="", block=""):
    """
    Args:
        orca (str): the ORCA executable
        simple (str): additional SimpleInput beside method
            and basis
        block (str): additional BlockInput beside ``pal``
            and ``maxcore``

    Per ASE's ORCA interface, the ORCA executable can also
    be set by envar ``ORCA_COMMAND``. This tasklet depends
    on OpenMPI >= 4.0 for parallelization.
    """
    m, b = mb.split('/')
    hdr = f"{m} {b} {simple}"
    blk = (
        f"%pal nprocs {GVAR.res.N} end\n"
        f"%maxcore {int(GVAR.res.m/GVAR.res.N*1024)}\n" # per core mem in MB
        f"{block}"
    )
    calc = ORCA(
        **__dl(cid), orcasimpleinput=hdr, orcablocks=blk,
        charge=chg_mlt[0], mult=chg_mlt[1],
    )
    ORCA_CMD = orca or environ.get('ORCA_COMMAND', 'orca')
    cpu = __cs(GVAR.res.c)
    calc.command = f'{ORCA_CMD} PREFIX.inp "--bind-to cpulist:ordered --cpu-set {cpu}" > PREFIX.out'
    return await ase_fio_task(cid, calc, atcor, ian)


def __dl(cid):
    cid = str(cid)
    if '/' not in cid: cid = "./"+cid
    d, l = cid.rsplit('/',1)
    return dict(directory=d, label=l)
def __cs(s):
    return ','.join(map(str,sorted(s)))



import setuptools

exec(open("ase_grain/_version.py").read())

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="ase-grain",
    version=__version__,
    author="Harry Zhang",
    author_email="zhanghar@iu.edu",
    description="An async wrapper for ASE",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Contextualist/ase-grain",
    packages=setuptools.find_packages(),
    install_requires=[
        "ase >= 3.22.0, < 3.23.0",
        "grain-scheduler >= 0.12.1",
        "trio",
    ],
    tests_require=[
        "pytest",
        "pytest-trio",
    ],
    python_requires=">=3.6",
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Framework :: Trio",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
    ],
)
